% Magic Square
% Guillaume Arnaud, INRIA Rocquencourt
% Date : 01/06/2003

ms(L,N):-
	% Data
	N2 is N*N,
	length(L,N2),
	fd_domain(L,1,N2),
	Som is N*(N2+1)//2,

	% All squares have a different value
	fd_all_different(L),
	
	% Sum constraints
	main_const(L,N,Som),

	sym_hor(N,SymHor),
	sym_rot(N,SymRot),
%	sym_val(N2,SymVal),
%	sym_alg1(SymAlg1),
	
	% Symmetry breaking labeling
	fd_labeling_sbds(L,[group([SymHor,SymRot])],[val_order(middle)]).
	

%%%%%%
% Sum constraints
main_const(L,N,Som):-
	hor_const(L,1,N,Som),
	ver_const(L,1,N,Som),
	get_diag1(L,N,1,Diag1),
	sum_const(Diag1,Som,Som),
	get_diag2(L,N,N,Diag2),
	sum_const(Diag2,Som,Som).
	
% hor constraints
hor_const(_,I,N,_):-I>N,!.
hor_const(L,I,N,Som):-
	get_line(L,N,I,LI),
	sum_const(LI,Som,Som),
	Ip1 is I+1,
	hor_const(L,Ip1,N,Som).
	
% ver constraints
ver_const(_,I,N,_):-I>N,!.
ver_const(L,I,N,Som):-
	get_row(L,N,I,LI),
	sum_const(LI,Som,Som),
	Ip1 is I+1,
	ver_const(L,Ip1,N,Som).


%%%%%%
% get Ith line of L
get_line(L,N,I,Lout):-
	R1 is (I-1)*N,
	RI is N*N-R1-N,
	length(L1,R1),
	length(Lout,N),
	length(L2,RI),
	flatten([L1,Lout,L2],L).
	
% get Ith row of L
get_row(L,N,I,Lout):-
	length(Lout,N),
	get_row_aux(N,I,L,Lout).
	
get_row_aux(_,_,_,[]).
get_row_aux(N,I,Lin,[X|L]):-
	nth(I,Lin,X),
	Inxt is I+N,
	get_row_aux(N,Inxt,Lin,L).
	
% get first diagonal
get_diag1(_,N,I,[]):-I>N*N,!.
get_diag1(L,N,I,[X|RL]):-
	nth(I,L,X),
	Inxt is I+N+1,
	get_diag1(L,N,Inxt,RL).
	
% get first diagonal
get_diag2(_,N,I,[]):-I>=N*N,!.
get_diag2(L,N,I,[X|RL]):-
	nth(I,L,X),
	Inxt is I+N-1,
	get_diag2(L,N,Inxt,RL).
	
% A+B+C+...#=S 
sum_const([A],A,_):-!.
sum_const([A|[B|R]],S,Sum):-
	fd_domain(S2,1,Sum),
	sum_const([B|R],S2,Sum),
	A+S2#=S.
	
% Horizontal symmetry (A11 -> A1N)
sym_hor(N,sym(varval,LVarSym,LVal)):-
	N2 is N*N,
	liste_num(LVal,0,N2),
	liste_num(LVar,1,N2),
	build_hor_sym(LVar,N,LVarSym).

build_hor_sym([],_,[]):-!.
build_hor_sym(LVar,N,LVarOut):-
	length(Line,N),	
	append(Line,RLVar,LVar),
	reverse(Line,LineSym),
	build_hor_sym(RLVar,N,RLVarOut),
	append(LineSym,RLVarOut,LVarOut).	
	
% Rot +90 symmetry
sym_rot(N,sym(varval,LVarSym,LVal)):-
	N2 is N*N,
	liste_num(LVal,0,N2),
	length(LVarSym,N2),
	build_rot_sym(1,N,LVarSym).

	
build_rot_sym(_,_,[]):-!.
build_rot_sym(X,N,[XSym|L]):-
	I is ((X-1) mod N) +1,
	J is (X-1)//N+1,
	
	XSym is (I-1)*N+J,
	Xp1 is X+1,
	build_rot_sym(Xp1,N,L).

% Val symmetry
sym_val(N2,sym(varval,LVar,[0|LValSym])):-
	liste_num(LVar,1,N2),
	liste_num(LVal,1,N2),
	reverse(LVal,LValSym).


% Algebric symmetry 1 (for N=4)
sym_alg1(sym(_,[[2,3],[5,9],[8,12],[6,11],[7,10],[14,15]],[],red)).

%%%%%%
% Display
write_square([],_):-!.
write_square(L,N):-
	length(L1,N),
	append(L1,L2,L),
	write(L1),nl,
	write_square(L2,N).
