% Social Golfer problem
% http://4c.ucc.ie/~tw/csplib//prob/prob010/index.html

% April 30 2003: Version 0.1.1
% Pairwise representation, version for sbds_v7

golf3(G,S,W,R) :-
	% Data
	N is G*S,
	length(R,N),
	init_triangle(R,N,W),
	length(TriSInf,N),
	triangle_strict_inf(R,TriSInf,N),
	length(Square,N),
	triangle_to_square(TriSInf,R,Square),
	flatten(R,Vars),
	% Breaking symmetries statically
	nth(1,R,First_Player),First_Player=[_|RFirst_Player],
	ordering_weeks(RFirst_Player),
	% Constraint
	group_buildings(Square,R,W,S,N),
	% Redundant constraints
%	reverse(TriSInf,RStrictInf),
%	Wp1 is W+1,
%	plus_diag(RStrictInf,TriInf,Wp1),
%	transitivity_redund(TriInf),
	% Labeling
	randomize,	
	fd_labeling(Vars,[variable_method(first_fail),value_method(middle)]),
	convert_repr(Square,1,ListList,W,G),
	nl,nl,nl,write_sol(ListList).
golf_sbds(G,S,W,Square) :-
	% Data
	N is G*S,
	length(R,N),
	init_triangle(R,N,W),
	length(TriSInf,N),
	triangle_strict_inf(R,TriSInf,N),
	length(Square,N),
	triangle_to_square(TriSInf,R,Square),
	flatten(Square,LVars),

	% Constraint
	group_buildings(Square,R,W,S,N),
	
	% Redundant constraints
	NbGr is S*(S-1)*G // 2,
	flatten(TriSInf,TriSFlat),
	group_redund(TriSFlat,NbGr,W),

%	reverse(TriSInf,RStrictInf),
%	Wp1 is W+1,
%	plus_diag(RStrictInf,TriInf,Wp1),
%	transitivity_redund(TriInf),
	
	% Breaking symmetries statically
	nth(1,R,First_Player),First_Player=[_|RFirst_Player],
	ordering_weeks(RFirst_Player),
%	last(Square,Last_Player),
%	NoWithFirst is N-(S-1)*W-1,
%	length(LastLast,NoWithFirst),
%	append(_,LastLast,Last_Player),
%	reverse(LastLast,LastLastRev),
%	ordering_weeks(LastLastRev),	
%	Sm1 is S-1,
%	length(LAux,S),append(LAux,RR,R),
%	length(Gr1W2,Sm1),append(Gr1W2,_,RR),
	
	
	
	% SBDS Labeling
%	liste_num(LIVars,2,N),
	
	create_sym(LSym,G,S,W),
	length(LSym,NbSym),
	write('Nb de symtrie : '),write(NbSym),nl,nl,

%	LSym2=[],
	randomize,
	labeling_sbds(LVars,LSym,[val_order(random),var_order(standard)]),
%	fd_labeling(LVars,[value_method(random)]),
	convert_repr(Square,1,ListList,W,G),
	nl,nl,nl,write_sol(ListList),nl,
%	nth(2,Square,Sec),%write(Sec),nl,
	write_sq(Square),
	true.



init_triangle([],_,_).
init_triangle([P1|LP],N,W) :-
	length(LP,N2),
	LgP1 is N2+1,
	length(P1,LgP1),
	P1=[Id|LVar],
	fd_domain(LVar,0,W),
	Wp1 is W+1,
	fd_domain(Id,Wp1,Wp1),
	init_triangle(LP,N,W).

triangle_strict_inf(_,[],_).
triangle_strict_inf(R,[P1|TriInf],N) :-
	length(TriInf,LgRst),
	I is  N-LgRst,
	LgP1 is N-LgRst-1,
	length(P1,LgP1),
	tr_Tri(P1,R,1,I),
	triangle_strict_inf(R,TriInf,N).

tr_Tri([],_,_,_) :-!.
tr_Tri([X|LP1],R,J,I) :-
	J2 is J+1,
	I2 is I-1,
	nth(J,R,P2),
	nth(I,P2,X),
	tr_Tri(LP1,R,J2,I2).

triangle_to_square([],[],[]).
triangle_to_square([Pinf|LTriInf],[Psup|LPTriSup],[Psq|LPSq]) :-
	append(Pinf,Psup,Psq),
	triangle_to_square(LTriInf,LPTriSup,LPSq).
	
plus_diag([],[],_).	
plus_diag([P|RP],[P2|RP2],W):-
	reverse(P,P22),
	append([W],P22,P2),
	plus_diag(RP,RP2,W).

% Breaking Symmetries
ordering_weeks([]).
ordering_weeks([_]).
ordering_weeks([A|[B|R]]) :-
	A#>=B,
	ordering_weeks([B|R]).
%
last_week([],_,_,_,[]):-!.
last_week([P|LP],G,S,W,L) :-
	Sm1 is S-1,
	fix_P(P,S,Sm1,W,L1),
	last_week(LP,G,S,W,L2),
	append(L1,L2,L).

fix_P(_,_,0,_,[]):-!.
fix_P(P,S,Sm1,W,[X|R]):-
	I is S+(Sm1-1)*Sm1,
	nth(I,P,X),
%	write(I-P-X),nl,
%	X#=W,
	Sm2 is Sm1-1,
	fix_P(P,S,Sm2,W,R).

% Constraints
group_buildings([],[],_,_,_).
group_buildings([P1|LPlayer],[P11|LPlayerTri],W,S,N) :-
	Sm1 is S-1,
	NbZero is N-W*(S-1)-1,
	P11=[_|RP11],
	(NbZero=0->fd_domain(RP11,1,W);
		fd_exactly(NbZero,P1,0)),
	group_week(P1,W,Sm1),
	
	% Redundant constraints
	magicNumber(S,W,Sum1),
	compute_sum(P1,Sum2),	
	Sum2#=Sum1+W+1,

	transitivity(RP11,LPlayerTri),
	group_buildings(LPlayer,LPlayerTri,W,S,N).


group_week(_,0,_) :-!.
group_week(P1,W,S) :-
	Wm1 is W-1,	
	fd_exactly(S,P1,W),
	group_week(P1,Wm1,S).

transitivity([],_).
transitivity([_],_) :-!.
transitivity([V1|LV],[Player|LPlayer]) :-
	Player=[_|RPlayer],
	cond_const(V1,LV,RPlayer),
	transitivity(LV,LPlayer).

cond_const(_,[],_):-!.
cond_const(V1,[V2|RV],[V3|RPlayer]) :-
%	fd_cardinality([V1#>0,V2#>0,V3#>0],T1),
%	fd_domain(T2,[0,3]),
%	fd_cardinality([V1#=V2,V1#=V3,V2#=V3],T2),
%	(T1#=1)#==>T2#=0,
%	(T1#=0)#==>(T2#=0#\/T2#=3),

%	(#/\((V1#>0),(V1#=V2)))#==>(V3#=V1),
	fd_transitivity([V1,V2,V3]),

%	(V1#=V2)#==>(V3 #= V2),
%	(#/\((V1#>0),(V1#=V2)))#==>(#/\((V3 #= V2),V3#>0)),
%	(#/\((V1#>0)#/\(V2#>0),(V1#\=V2))) #==> (V3 #\= V2),
	cond_const(V1,RV,RPlayer).

% Redundant Constraints
compute_sum([], 0).
compute_sum([E|L], S):-
	compute_sum(L, S1),
	S #= S1 + E.
magicNumber(S, 1, S1):-
	!,
	S1 is S - 1.
magicNumber(S, W, MagicNumber):-
	W1 is W - 1,
	magicNumber(S, W1, MagicNumberW1),
	MagicNumber is MagicNumberW1 + W * (S - 1).
	

transitivity_redund([]).
transitivity_redund([P1|LPlayer]):-
	P1=[_|RP1],
	transitivity(RP1,LPlayer),
	transitivity_redund(LPlayer).

group_redund(_,_,0):-!.
group_redund(R,NbGr,W) :-
	fd_exactly(NbGr,R,W),
	Wm1 is W-1,
	group_redund(R,NbGr,Wm1).

% Adapted labeling
labeling_golf([_],_):-!.
labeling_golf([P|LP],K) :-
	P=[_|RP],
	randomize,
	labeling_player(RP,[],LP,1),
	Kp1 is K+1,
	labeling_golf(LP,Kp1).

labeling_first_player([],_,_,_).
labeling_first_player([V|LV],LV0,LP,J) :-
	(V>0->same_group(LV0,1,J,LP,V);true),
	Jp1 is J+1,
	append(LV0,[V],LV1),
	labeling_first_player(LV,LV1,LP,Jp1).
	
labeling_player([],_,_,_).
labeling_player([V|LV],LV0,LP,J) :-
%	write('labeling'-V),nl,
	fd_labeling(V,[value_method(random)]),
%	write(LV0-V),nl,
	(V>0->same_group(LV0,1,J,LP,V);true),
	Jp1 is J+1,
	append(LV0,[V],LV1),
	labeling_player(LV,LV1,LP,Jp1).
	
same_group([],_,_,_,_):-!.
same_group([V1|LV1],I,J,LP,V1) :-
	!,
	nth(I,LP,P2),
	JJ is J-I+1,
	nth(JJ,P2,V2),
	V2#=V1,
	Ip1 is I+1,
	same_group(LV1,Ip1,J,LP,V1).
same_group([_|LV1],I,J,LP,V2) :-
	nth(I,LP,P2),
	JJ is J-I+1,
	nth(JJ,P2,V3),
	V3#\=V2,
	Ip1 is I+1,
	same_group(LV1,Ip1,J,LP,V2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Symmetry					%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

create_sym(LSym,G,S,W):-
	N is G*S,
	Sm1 is S-1,
	NbVar is G*S*G*S,
	NbVal is W+2,
	liste_num(LIVars,2,N),
	sym_player(LIVars,N,W,Sm1,LSymPlayer,NbVar,NbVal),
%	sym_week(W,LSymWeek),
	LSymWeek=[],
%	LSymPlayer=[],
	flatten([LSymWeek,LSymPlayer],LSym).


% Symmetry constraint

sym_player([],_,_,_,[],_,_):-!.
sym_player(LInd,N,0,_,LSym,NbVar,NbVal):-
	!,
	sym_player_G(LInd,LInd,N,LSym,NbVar,NbVal).
sym_player(LInd,N,W,S,LSym,NbVar,NbVal):-
	length(LInd1,S),
	append(LInd1,LInd2,LInd),
	sym_player_G(LInd1,LInd1,N,LSym1,NbVar,NbVal),

	W2 is W-1,
	sym_player(LInd2,N,W2,S,LSym2,NbVar,NbVal),
	append(LSym1,LSym2,LSym).

sym_player_G([_],_,_,[],_,_):-!.
sym_player_G([P1|RLInd],LInd,N,LSym,NbVar,NbVal):-
	sym_player_P1(P1,RLInd,N,LSym2,LInd,NbVar,NbVal),
	sym_player_G(RLInd,LInd,N,LSym1,NbVar,NbVal),
	append(LSym1,LSym2,LSym).

sym_player_P1(_,[],_,[],_,_,_).
sym_player_P1(P1,[P2|RLInd],N,[Sym|LSym],LInd,NbVar,NbVal):-
	sym_player_P1_P2(P1,P2,N,Sym,NbVar,NbVal),
	sym_player_P1(P1,RLInd,N,LSym,LInd,NbVar,NbVal).
	
sym_player_P1_P2(P1,P2,N,sym(Sym,LPoint),NbVar,NbVal):-
	number_atom(P1,SuffixeP1),
	number_atom(P2,SuffixeP2),
	atom_concat('sym_var_',SuffixeP1,Prefixe1),
	atom_concat(Prefixe1,'_',Prefixe2),
	atom_concat(Prefixe2,SuffixeP2,Sym),
	write(Sym),nl,
	
	list_of_list(LX,N,N),
	transpo_list_of_list(LX,LY),
	nth(P1,LX,P1X),
	retirer_liste(P1,P1X,L1Aux),P2m1 is P2-1,retirer_liste(P2m1,L1Aux,LP1XAux),
	nth(P2,LX,P2X),
	retirer_liste(P1,P2X,L2Aux),retirer_liste(P2m1,L2Aux,LP2XAux),
	nth(P1,LY,P1Y),
	retirer_liste(P1,P1Y,L3Aux),retirer_liste(P2m1,L3Aux,LP1YAux),
	nth(P2,LY,P2Y),
	retirer_liste(P1,P2Y,L4Aux),retirer_liste(P2m1,L4Aux,LP2YAux),

	append(LP1YAux,LP1XAux,LP1),
	append(LP2YAux,LP2XAux,LP2),
	
%	subtract(LI,LOrb,LStab),
	NbAff is NbVar*NbVal,
	length(LPoint,NbAff),
	var_to_point(LP1,LP2,1,LPoint,NbVar,NbVal) .



var_to_point(_,_,_,[],_,_).
%var_to_point(N,M,X,[X|L],NbVar,_):-
%	X>NbVar,!,Xp1 is X+1,var_to_point(N,M,Xp1,L,NbVar,_),
%	write('titi'-X),nl.
var_to_point(L1,L2,X,[XSym|LPoint],NbVar,NbVal):-
	I is (X-1)//NbVal+1,
	J is (X-1) mod NbVal,
	(nth(X1,L1,I)->
		nth(X1,L2,X2),
		XSym is (X2-1)*NbVal+J+1;
		(nth(X2,L2,I)->
		nth(X2,L1,X1),
		XSym is (X1-1)*NbVal+J+1;
		XSym=X)),
	Xp1 is X+1,
	var_to_point(L1,L2,Xp1,LPoint,NbVar,NbVal).

list_of_list(L,X,Y):-
	length(L,X),
	enr_list(L,1,Y,Y).
enr_list([],_,_,_).
enr_list([L1|RL],Y1,Y2,Y):-
	length(L1,Y),
	liste_num(L1,Y1,Y2),
	Y1p is Y1+Y,
	Y2p is Y2+Y,
	enr_list(RL,Y1p,Y2p,Y).

transpo_list_of_list(L1,L2):-
	length(L1,X),
	length(L2,X),
	enr_transpo(L1,L2,1),!.

enr_transpo(_,[],_).
enr_transpo(L1,[LL2|L2],X):-	
	findall(I,(member(LL1,L1),nth(X,LL1,I)),LL2),
	Xp1 is X+1,
	enr_transpo(L1,L2,Xp1).

% Symtries semaines

sym_week(W,LSym) :-
	sym_week_W1(1,2,W,LSym).

sym_week_W1(W,_,W,[]):-!.
sym_week_W1(W1,W2,W,LSym) :-
	sym_week_W1_W2(W1,W2,W,LSym2),
	W1p1 is W1+1,
	W2p1 is W2+1,
	sym_week_W1(W1p1,W2p1,W,LSym1),
	append(LSym1,LSym2,LSym).

sym_week_W1_W2(_,W2,W,[]) :- W2>W,!.
sym_week_W1_W2(W1,W2,W,[sym(Sym,(W1,W2),val)|LSym]) :-
	number_atom(W1,SuffixeW1),
	number_atom(W2,SuffixeW2),
	atom_concat('sym_week_',SuffixeW1,Prefixe1),
	atom_concat(Prefixe1,'_',Prefixe2),
	atom_concat(Prefixe2,SuffixeW2,Sym),
	
%	liste_num(LVar,1,NbVar),
%	NbVarp1 is NbVar+1,
%	NbValMax is NbVal+NbVar,
%	W1pVar is W1+NbVar+1,
%	W2pVar is W2+NbVar+1,
%	liste_num(LVal,NbVarp1,NbValMax),
%	permute_liste(LVal,W1pVar,W2pVar,LValPermute),
%	append(LVar,LValPermute,LPoint),
	W2p1 is W2+1,
	sym_week_W1_W2(W1,W2p1,W,LSym)	
	.

permute_liste([],_,_,[]).
permute_liste([X|L],X,Y,[Y|L2]):-!,
	permute_liste(L,X,Y,L2).
permute_liste([Y|L],X,Y,[X|L2]):-!,
	permute_liste(L,X,Y,L2).
permute_liste([V|L],X,Y,[V|L2]):-
	permute_liste(L,X,Y,L2).

% Symtries composes (pour (5,3,7) uniquement)
sym_comp(LSym,SymStream):-
	L=[7-(2,3),6-(4,5),5-(6,7),4-(8,9),3-(10,11),2-(12,13),1-(14,15)],
	write_sym_comp(L,LSym,SymStream).


write_sym_comp([_],[],_):-!.
write_sym_comp([W-(X1,X2)|Rest],LSym,SymStream):-
	write_sym_comp_suite(W,X1,X2,Rest,LSym1,SymStream),
	write_sym_comp(Rest,LSym2,SymStream),
	append(LSym1,LSym2,LSym).
	
write_sym_comp_suite(_,_,_,[],[],_):-!.
write_sym_comp_suite(WX,X1,X2,[WY-(Y1,Y2)|Rest],[Sym|LSym],SymStream) :-
	number_atom(X1,Suffixe1),
	number_atom(X2,Suffixe2),
	number_atom(Y1,Suffixe3),
	number_atom(Y2,Suffixe4),
	atom_concat('sym_comp_',Suffixe1,Prefixe1),
	atom_concat(Prefixe1,Suffixe2,Prefixe2),
	atom_concat(Prefixe2,Suffixe3,Prefixe3),
	atom_concat(Prefixe3,Suffixe4,Sym),
	Diff is Y1-X1,

	format(SymStream,
		'~a(_,_,I1,Val,I2,Val):-\n\t I1>18,\n\t ~d is I1 mod 15,\n\t I2 is I1+~d.\n', [Sym,X1,Diff]),
	format(SymStream,
		'~a(_,_,I1,Val,I2,Val):-\n\t I1>18,\n\t ~d is I1 mod 15,\n\t I2 is I1+~d.\n', [Sym,X2,Diff]),
	format(SymStream,
		'~a(_,_,I1,Val,I2,Val):-\n\t I1>18,\n\t ~d is I1 mod 15,\n\t I2 is I1-~d.\n', [Sym,Y1,Diff]),
	format(SymStream,
		'~a(_,_,I1,Val,I2,Val):-\n\t I1>18,\n\t ~d is I1 mod 15,\n\t I2 is I1-~d.\n', [Sym,Y2,Diff]),
	format(SymStream,
		'~a(_,_,I,~d,I,~d):-\n\t I>18.\n',[Sym,WX,WY]),
	format(SymStream,
		'~a(_,_,I,~d,I,~d):-\n\t I>18.\n',[Sym,WY,WX]),
%%%
	format(SymStream,
		'~a(_,_,I,Val,I,Val):-\n\t I < 16,!.\n', [Sym]),		
	format(SymStream,
		'~a(_,_,I,Val,I,Val):-\n\t 1 is I mod 15,!.\n', [Sym]),		
%	format(SymStream,
%		'~a(_,_,I1,Val1,I2,Val2):-\n\t ~d is I mod 15,\n\t, A is (I1-1)//15+1,\n\t,A\=~d, A\=~d,\n\t I2 is I1+~d,(Val1=WX->Val2=WY).\n', [Sym,X1,X1,Y1]),
	write_sym_comp_suite(WX,X1,X2,Rest,LSym,SymStream).
	
% Display
write_sol([]).
write_sol([P|LP]):-
	write_week(P),
	nl,
	write_sol(LP).

write_week([]).
write_week([Group|LG]):-
	write_group(Group),
	write(' |'),
	write_week(LG).

write_group([]).
write_group([I|R]):-
	write(I),write(' '),
	write_group(R).
	
convert_repr([],_,L,W,_):-
	length(L,W),
	init_list(L,[]),
	!.
convert_repr([P|LP],I,L2,W,G):-
	Ip1 is I+1,
	convert_repr(LP,Ip1,L,W,G),
	group_week(P,I,W,LG),
	rec_weeks(LG,L,W,L2).

group_week(_,_,0,[]):-!.
group_week(P,I,W,LG):-
	setof(IG,nth(IG,P,W),G),
	sort([I|G],Group1),
	Wm1 is W-1,
	group_week(P,I,Wm1,LG2),
	append([Group1],LG2,LG).
	
rec_weeks([],L,0,L).
rec_weeks([LGW|LG],L,W,L3):-
	Wm1 is W-1,
	rec_weeks(LG,L,Wm1,L2),
	nth(W,L,Week),
	(member(LGW,Week)->
		L3=L2;
		append(Week,[LGW],Week2),
		sort(Week2,Week3),
		replace_list(W,L2,Week3,L3)
	).
		
write_sq([]).
write_sq([Lg|R]):-
	write(Lg),nl,
	write_sq(R).

% Result:

% golf3(5,3,7,R):
% portnb(newVariable, 1940).
% portnb(newConstraint, 2618).
% nbcstr(6389).
