% Version 8
% 24/04/2003
% Last changed : 23/05/2003

% Non expression des listes de permutation en prolog (optimisation des tailles de piles)
% Integration de la gnration de groupe (algorithme Dimino)
% Les deux reprsentations possibles

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Fonction de labeling utilisant la SBDS %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Structure de donnes			 %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
					
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Liste de variables:

% liste_num(Lindex,I,N) (voir le fichier liste.pl) cre une 
% liste de nombres entiers successifs de I  N, cette liste
% correspond aux indices des variables et sera utilise pour
% dterminer les symtries.

fd_labeling_sbds(L,S):-
	init_sym(L,S,[],LIVar,N,STot,ValMeth,LVar),
	sbds(LIVar,L-N-STot,[],value_method(ValMeth),LVar).
fd_labeling_sbds(L,S,Param):-
	init_sym(L,S,Param,LIVar,N,STot,ValMeth,LVar),
%	length(STot,Lg),write(Lg),nl,
	sbds(LIVar,L-N-STot,[],value_method(ValMeth),LVar).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Corps du programme			 %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% init_sym(+L,+S,+Param) :
% fonction principale de labeling, appele par l'utilisateur
% au moment de l'numration (remplace fd_labeling).

% L= liste des variables,
% S= Liste des symtries,

init_sym(L,S):-
	length(L,N),
	liste_num(L_IVar,1,N),
	Val_order=min,
	sbds(L_IVar,L-N-S,[],value_method(Val_order),L).

% Param= Liste des paramtres  passer pour dfinir les symtries.

init_sym(L,S):-
	length(L,N),
	liste_num(L_IVar,1,N),
	Val_order=min,
	sbds(L_IVar,L-N-S,[],value_method(Val_order),L).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% init_sym(+L,+S,+Label_method,+Param) :
% fonction semblable  init_sym/2, mais avec en plus
% la gestion des mthodes d'numrations contenues dans
% Label_method.

% Label_method= var_order(Var_order),value_method(Val_order)...
% Les mthodes existantes sont :
% Var_order= 	normal (on balaie les variables de gauche  droite),
%		middle (on commence l'numration par le milieu),
%		extrem (on commence l'numration par les extrmes);
% Val_order=	mme chose que pour Var_order.
%%% A faire= first_fail, most_constrainted.... %%%

value_choice_pl([min,max,middle,bounds,random]).

init_sym(L,S,Label_method,L_IVar5,N,STot,ValMeth,LVar5):-
	length(L,N),
	liste_num(L_index,1,N),
	g_assign(lab_meth,[]),
	(member(variable_method(Var_order),Label_method)->
		(Var_order=middle
			->
			list_sort_middle(L,L_reorganized),
			list_sort_middle(L_index,L_IVar),
			true;
		(Var_order=bounds
			->
			list_sort_extrem(L,L_reorganized),
			list_sort_extrem(L_index,L_IVar),
			true;
		(Var_order=max
			->
			reverse(L,L_reorganized),
			reverse(L_index,L_IVar);
		(Var_order=standard
			->
			L_IVar=L_index,
			L_reorganized=L,	
			g_assign(lab_meth,[variable_method(Var_order)])
			;			
			L_IVar=Var_order,
			list_sort(L,L_reorganized,Var_order)
		))
		)
		)
		;
	L_IVar=L_index,
	L_reorganized=L),	
	
	(member(value_method(ValMeth),Label_method)->
%		g_read(lab_meth,Lab_meth),
%		g_assign(lab_meth,[value_method(Val_order)|Lab_meth]),
		true;
		ValMeth=min
	),

	which_rep(S,Rep),
%	write(Rep),nl,
	(Rep=1	->
		asserta((image_sym(Sym,(IVar,Val),(IVarSym,ValSym)):-!,image_sym_c_1(Sym,IVar,Val,IVarSym,ValSym)))
		;
		asserta((image_sym(Sym,(IVar,Val),(IVarSym,ValSym)):-!,image_sym_c_2(Sym,IVar,Val,IVarSym,ValSym)))),
	initialisation_sym_pl(L,S,STot,Rep),
%	value_choice_pl(LValMeth),nth(ValMeth,LValMeth,Val_order),

	reorder_inst(L_IVar,L_reorganized,L_IVar2,LVar2,L_IVar3,LVar3),
	append(L_IVar3,L_IVar2,L_IVar5),
	append(LVar3,LVar2,LVar5).

which_rep([],2):-!.
%which_rep([group(_)|_],1):-!.
which_rep([sym(point,_)|_],1):-!.
which_rep([sym(val_local,_,_)|_],1):-!.
which_rep([sym(var_local,_,_)|_],1):-!.
which_rep([group(S)|LSym],Rep):-!,append(S,LSym,NewSym),which_rep(NewSym,Rep).
which_rep([_|LSym],Rep):-which_rep(LSym,Rep).
	

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% reorder_inst()

reorder_inst([],[],[],[],[],[]):-!.
reorder_inst([IVar|LIVar],[Var|LVar],[IVar|LIVar2],[Var|LVar2],LIVar3,LVar3):-
	fd_size(Var,Size),Size>1,!,
	reorder_inst(LIVar,LVar,LIVar2,LVar2,LIVar3,LVar3).
reorder_inst([IVar|LIVar],[Var|LVar],LIVar2,LVar2,[IVar|LIVar3],[Var|LVar3]):-
	reorder_inst(LIVar,LVar,LIVar2,LVar2,LIVar3,LVar3).
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% sbds_init(+L_IVar,+S,+A,+Val_method,) 
sbds_init([],_,A,A,[],[]):-!.
sbds_init(L_IVar,S,A,A_out,L_reorganized,Sinit):-
	S=L-N-_,
	L_reorganized=[Var|L_reorganized2],
	L_IVar=[IVar|L2],
	
	% Noeud de l'arbre de recherche
	(fd_dom(Var,[Val])	->
				Var#=Val,
				g_assign(sred,[]),
				\+reduction_ens_sym(S,IVar,Val,A),
				g_read(sred,Sred),
				append(Sred,Sinit2,Sinit),
				sbds_init(L2,L-N-Sred,[IVar|A],A_out,L_reorganized2,Sinit2)
				;
				sbds_init(L2,S,A,A_out,L_reorganized2,Sinit)
	).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% sbds(+L_IVar,+S,+A,+Val_method,+Param) :
% construit l'arbre de recherche.

% L_IVar= liste des indices des variables non encore numres;
% S= de la forme LVar-N-LSym, o LVar est la liste des variables;
%    N le nombre de variables et LSym la liste des symtries;
% A= liste des indices des variables dj affectes;
% Val_method= mthode d'numration des valeurs.

sbds([],_,_,_,[]):-!.
sbds(_,_-_-[],_,Options,[Var|L_reorganized]) :-
	!,
%	g_read(lab_meth,Lab_meth),
	fd_labeling(Var,[Options]),
	sbds(_,_-_-[],_,Options,L_reorganized).
sbds(L_IVar,S,A,value_method(Val_order),L_reorganized):-
	S=L-N-_,
%	choix_variable(L_reorganized,L_IVar,Var,IVar,L2,L_reorganized2),
	L_reorganized=[Var|L_reorganized2],
	L_IVar=[IVar|L2],
	
	fd_dom(Var,Dom),
%	fd_size(Var,Size),
%	value_choice_c(Val_order,Size,ValInd),
%	nth(ValInd,Dom,Val),
	fd_domain(VarAux,Dom),
	fd_labeling(VarAux,[value_method(Val_order)]),!,
	fd_dom(VarAux,[Val]),

	% Noeud de l'arbre de recherche
	(
		% Affectation
		Var#=Val,
		% sred est une variable globale contenant les symtries restantes
		% du problme
		g_assign(sred,[]),
		\+reduction_ens_sym(S,IVar,Val,A),
		g_read(sred,Sred),
		(member(IVar,A)->NewA = A;NewA = [IVar|A]),
		sbds(L2,L-N-Sred,NewA,value_method(Val_order),L_reorganized2)
		;
		% Rduction du domaine de recherche
		Var#\=Val,
		% Ajout des contraintes de symtries
		(A=[]-> 
			retirer_affectation_sym(IVar,Val,S)
			;
			retirer_affectation_sym_cond(IVar,Val,S,A)
		),
		sbds(L_IVar,S,A,value_method(Val_order),L_reorganized)
	)
   	
	.
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% choix_variable(+List_var,+L1,-Var,-IVar,-L2) :
% choisi une variable Var  la position IVar dans la liste L,
% met  jour la liste L2 des variables non affectes.
choix_variable([],[],_,_,_,_).
choix_variable([Var|L_reorg2],[IVar|L2],Var,IVar,L2,L_reorg2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% reduction_ens_sym(+S,+Sred,+IVar,+Val,+Param) :
% Recherche et limination des symtries de l'affectation de la 
% variable d'indice IVar  la valeur Val, non consistantes
% avec l'affectation courante.
%
% S : Ensemble des symtries courantes
% Sred : Nouvel ensemble des symtries

%reduction_ens_sym(_-_-[],_,_,_,_):-fail.
reduction_ens_sym(L-N-LSym,IVar,Val,LIVar):-
	member(sym(Sym,_),LSym),
	g_read(sred,Sred),
	(image_sym(Sym,(IVar,Val),(IVarSym_t,ValSym_t))->
		nth(IVarSym_t,L,VarSym),
		(
		(VarSym#=ValSym_t, aff_sym(L-N-Sym,LIVar))
			->
			g_assign(sred,[sym(Sym,_)|Sred])
			;fail
		)
		;
		g_assign(sred,[sym(Sym,_)|Sred])),
	fail.
reduction_ens_sym2(L-_-LSym,IVar,Val,_):-
	member(sym(Sym,_),LSym),
	g_read(sred,Sred),
	(image_sym(Sym,(IVar,Val),(IVarSym_t,ValSym_t))->
		nth(IVarSym_t,L,VarSym),
		(
		(VarSym#=ValSym_t)
			->
			g_assign(sred,[sym(Sym,_)|Sred])
			
			;
			fail
		)
		;
		g_assign(sred,[sym(Sym,_)|Sred])),
	fail.
reduction_ens_sym3(L-_-LSym,IVar,Val,_):-
	member(sym(Sym,_),LSym),
	g_read(sred,Sred),
	image_sym(Sym,(IVar,Val),(IVarSym_t,ValSym_t)),
	nth(IVarSym_t,L,VarSym),
	fd_dom(VarSym,Dom),
	(memberchk(ValSym_t,Dom) ->
		g_assign(sred,[sym(Sym,_)|Sred])
		;fail
	),
	fail.
% aff_sym(+L-N-Sym,+A,+Param) :
% Pose les contraintes symtriques des affectations contenues
% dans A.

aff_sym(_-_-_,[]).
aff_sym(L-N-Sym,[IVar|A]):-
	nth(IVar,L,Val),					
	(image_sym(Sym,(IVar,Val),(IVarSym_t,ValSym_t))->
		IVarSym=IVarSym_t,ValSym=ValSym_t,
		nth(IVarSym,L,VarSym),
		VarSym#=ValSym;
		true
	),
	aff_sym(L-N-Sym,A).
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% retirer_affectation_sym(+IVar,+Val,+S,+Param) : 
% Retirer les affectations symtriques de Var->Val
% en ajoutant les contraintes de symtrie correspondantes

retirer_affectation_sym(_,_,_-_-[]).
retirer_affectation_sym(IVar,Val,L-N-[sym(Sym,_)|LSym]):-
	(image_sym(Sym,(IVar,Val),(IVarSym_t,ValSym_t))->
		nth(IVarSym_t,L,VarSym),
		VarSym#\=ValSym_t
		;
		true),
	retirer_affectation_sym(IVar,Val,L-N-LSym).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% retirer_affectation_sym_cond(+IVar,+Val,+S,+A,+Nb_aff,+Param) :
% Ajout de contraintes de symtries conditionnes par les
% affectations symtriques de l'affectation courante (A).

% Nb_aff= le nombre d'affectation courante, ncessaire pour
% poser la contrainte conditionnelle

retirer_affectation_sym_cond(_,_,_-_-[],_).
retirer_affectation_sym_cond(IVar,Val,L-N-[sym(Sym,_)|LSym],A):-
	(image_sym(Sym,(IVar,Val),(IVarSym_t,ValSym_t))->
		nth(IVarSym_t,L,VarSym),
		ValSym=ValSym_t,
		(((IVar,Val)=(IVarSym_t,ValSym_t))->Flag_Id1=1;Flag_Id1=0)
		;
		nth(IVar,L,VarSym),
		ValSym=Val,
		Flag_Id1=1 % L'affectation est invariante selon Sym
		),
	liste_aff_sym(L-N-Sym,A,ASym),
	length(ASym,LgCtes),
	(Flag_Id1=0->
		(LgCtes>0->
			fd_domain(Nb_cont_aff,0,LgCtes),
			fd_cardinality(ASym,Nb_cont_aff),
			(Nb_cont_aff#=LgCtes)#==>(VarSym#\=ValSym)
			;
			VarSym#\=ValSym	)
		;
		true),
	retirer_affectation_sym_cond(IVar,Val,L-N-LSym,A).

	
% liste_aff_sym(+S,+A,-ASym,+Param) :
% liste des affectations symtriques de A par rapport  Sym,
% sous la forme ASym=[VarSym#=ValSym,....,].
% S= L-N-Sym o Sym est une symtrie, L la liste des variables,
% N le nombre de variables.

liste_aff_sym(_-_-_,[],[]).
liste_aff_sym(L-N-Sym,[IVar|A],[VarSym#=ValSym_t|ASym]):-
	nth(IVar,L,Val),
	image_sym(Sym,(IVar,Val),(IVarSym_t,ValSym_t)),
		(IVar \= IVarSym_t;Val \= ValSym_t),
			nth(IVarSym_t,L,VarSym),
			liste_aff_sym(L-N-Sym,A,ASym),!
	.
liste_aff_sym(L-N-Sym,[_|A],ASym):-
	liste_aff_sym(L-N-Sym,A,ASym).
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Gestion des symtries					%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Calcule des couples (var,val) symtriques
%
% image_sym(+Sym,(+IVar,+Val),(-IVarSym,-ValSym)) :
% Sym = le nom du prdicat de symtrie
% IVar= l'indice de la variable
% Val = la valeur que prend la variable d'indice IVar
% (IVarSym,ValSym)= le couple image par la symtrie Sym de (IVar,Val)
:-dynamic(image_sym/3).
image_sym(_,_,_).
%image_sym(Sym,(IVar,Val),(IVarSym,ValSym)) :-
%	convert_sym(Sym,NumSym),
%	image_sym_c(Sym,IVar,Val,IVarSym,ValSym),!.



convert_sym(Sym,NumSym) :-
	g_read(lst_sym,L),
	member(Sym-NumSym,L).
	
% generate_pred_sym(+Liste_LPoints,-Liste_Pred_Sym,+NumSym,+Rep)
generate_pred_sym([],[],_,_):-!.
generate_pred_sym([LPoints|RPoints],[sym(Num,LPoints)|RSym],Num,1):-
	Nump1 is Num+1,
	generate_pred_sym(RPoints,RSym,Nump1,1).
generate_pred_sym([(LIVar,LVal)|RL],[sym(Num,LIVar,LVal)|RSym],Num,2):-
	Nump1 is Num+1,
	generate_pred_sym(RL,RSym,Nump1,2).
	
generate_pred_sym2(NbSym,[],NbSym):-!.
generate_pred_sym2(NbSym,[sym(Num,_)|RSym],Num):-
	Nump1 is Num+1,
	generate_pred_sym2(NbSym,RSym,Nump1).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Initialisation des symtries
%

% initialisation_sym_pl(+LVar,+LSym,-LSymOut,+Rep)

initialisation_sym_pl(LVar,LSym,LSymOut,Rep):-
	length(LVar,Nb_var),
	g_assign(nbvar,Nb_var),
	max_dom(LVar,0,Nb_valm1),
	Nb_val is Nb_valm1+1,
	g_assign(nbval,Nb_val),
	NbPoint is Nb_var*Nb_val,
	g_assign(nbpoint,NbPoint),
	length(LSym,Nb_sym),
	initialisation_sym_c(Nb_var,Nb_val,Nb_sym),
	initialisation_image_sym(LSym,1,LSymOut,Nb_var,Nb_val,Rep).

	
% initialisation_image_sym(+LSym,+NumSym,-LSymOut,+NbVar,+NbVal,+Rep)
initialisation_image_sym([],_,[],_,_,_).
initialisation_image_sym([sym(point,LPoint)|LSym],NumSym,[sym(NumSym,_)|LSymOut],Nb_var,Nb_val,Rep):-
	init_pl(NumSym,LPoint),
	NumSymp1 is NumSym+1,
	initialisation_image_sym(LSym,NumSymp1,LSymOut,Nb_var,Nb_val,Rep).
initialisation_image_sym([sym(val,(Val1,Val2))|LSym],NumSym,[sym(NumSym,_)|LSymOut2],Nb_var,Nb_val,Rep):-
	init_sym_val(LPerm,Val1,Val2,Nb_var,Nb_val,Rep),
	init_pl(NumSym,LPerm),
	NumSymp1 is NumSym+1,
	initialisation_image_sym(LSym,NumSymp1,LSymOut2,Nb_var,Nb_val,Rep).
initialisation_image_sym([sym(var,(Var1,Var2))|LSym],NumSym,[sym(NumSym,_)|LSymOut2],Nb_var,Nb_val,Rep):-
	init_sym_var(LPerm,Var1,Var2,Nb_var,Nb_val,Rep),
	init_pl(NumSym,LPerm),
	NumSymp1 is NumSym+1,
	initialisation_image_sym(LSym,NumSymp1,LSymOut2,Nb_var,Nb_val,Rep).
initialisation_image_sym([sym(var_local,(Var1,Var2),LVal)|LSym],NumSym,[sym(NumSym,_)|LSymOut2],Nb_var,Nb_val,Rep):-
	init_sym_var_local(LVal,Var1,Var2,Nb_var,Nb_val,LPoint),
	init_pl(NumSym,LPoint),
	NumSymp1 is NumSym+1,
	initialisation_image_sym(LSym,NumSymp1,LSymOut2,Nb_var,Nb_val,Rep).
initialisation_image_sym([sym(val_local,(Val1,Val2),LVar)|LSym],NumSym,[sym(NumSym,_)|LSymOut2],Nb_var,Nb_val,Rep):-
	init_sym_val_local(LVar,Val1,Val2,Nb_var,Nb_val,LPoint),
	init_pl(NumSym,LPoint),
	NumSymp1 is NumSym+1,
	initialisation_image_sym(LSym,NumSymp1,LSymOut2,Nb_var,Nb_val,Rep).
% init sym(_,[..],[..])
initialisation_image_sym([sym(red,LVar,LVal)|LSym],NumSym,[sym(NumSym,_)|LSymOut2],Nb_var,Nb_val,Rep):-
	init_sym_red_form(LPerm,LVar,LVal,Nb_var,Nb_val,Rep),
	init_pl(NumSym,LPerm),
	NumSymp1 is NumSym+1,
	initialisation_image_sym(LSym,NumSymp1,LSymOut2,Nb_var,Nb_val,Rep).
initialisation_image_sym([sym(varval,LVar,LVal)|LSym],NumSym,[sym(NumSym,_)|LSymOut2],Nb_var,Nb_val,Rep):-
	init_sym_var_val(LPerm,LVar,LVal,Nb_var,Nb_val,Rep),
	init_pl(NumSym,LPerm),
	NumSymp1 is NumSym+1,
	initialisation_image_sym(LSym,NumSymp1,LSymOut2,Nb_var,Nb_val,Rep).
% init group
initialisation_image_sym([group(LSym)|RLSym],NumSym,LSymOut,Nb_var,Nb_val,Rep):-
	write(LSym-Nb_var-Nb_val),nl,
	(Rep=2	->
		findall(Y,(member(Sym,LSym),Sym=sym(varval,LVar,LVal),Y=(LVar,LVal)),LSymInit1),!,
		findall(Y,(member(Sym,LSym),Sym=sym(var,(Var1,Var2)),init_sym_var(Y,NumSym,Var1,Var2,Nb_var,Nb_val,2)),LSymInit2),
		findall(Y,(member(Sym,LSym),Sym=sym(val,(Vall,Val2)),init_sym_val(Y,Val1,Val2,Nb_var,Nb_val,2)),LSymInit3),
		findall(Y,(member(Sym,LSym),Sym=sym(red,LVar,LVal),init_sym_red_form(Y,LVar,LVal,Nb_var,Nb_val,2)),LSymInit4),
		flatten([LSymInit1,LSymInit2,LSymInit3,LSymInit4],LSymInit)
		;
		findall(Y,(member(Sym,LSym),Sym=sym(point,Y)),LSymInit0),
		findall(Y,(member(Sym,LSym),Sym=sym(varval,LVar,LVal),convert_varval_to_point_list(Y,LVar,LVal,Nb_var,Nb_val)),LSymInit1),!,
		findall(Y,(member(Sym,LSym),Sym=sym(var,(Var1,Var2)),init_sym_var(Y,NumSym,Var1,Var2,Nb_var,Nb_val,1)),LSymInit2),
		findall(Y,(member(Sym,LSym),Sym=sym(val,(Vall,Val2)),init_sym_val(Y,Val1,Val2,Nb_var,Nb_val,1)),LSymInit3),
		findall(Y,(member(Sym,LSym),Sym=sym(var_local,(Var1,Var2),LVal),init_sym_var_local(LVal,Var1,Var2,Nb_var,Nb_val,Y)),LSymInit4),
		findall(Y,(member(Sym,LSym),Sym=sym(val_local,(Val1,Val2),LVar),init_sym_val_local(LVar,Val1,Val2,Nb_var,Nb_val,Y)),LSymInit5),
		findall(Y,(member(Sym,LSym),Sym=sym(red,LVar,LVal),init_sym_red_form(Y,LVar,LVal,Nb_var,Nb_val,1)),LSymInit6),
		flatten([LSymInit0,LSymInit1,LSymInit2,LSymInit3,LSymInit4,LSymInit5,LSymInit6],LSymInit)
	),
		
	NumSymm1 is NumSym-1,
	g_assign(numsym_cour,NumSymm1),
	write(Rep-LSymInit),nl,
	dimino(LSymInit,NbSymGen,Rep),!,	
	NewNumSym is NumSym+NbSymGen,
	generate_pred_sym2(NewNumSym,LSymOut1,1),
	initialisation_image_sym(RLSym,NewNumSym,LSymOut2,Nb_var,Nb_val,Rep),
	append(LSymOut1,LSymOut2,LSymOut).

% init list of points -> array in C
init_all_point([]).
init_all_point([sym(NumSym,LPoints)|LSym]):-
	init_point_pl(NumSym,LPoints,1),
	init_all_point(LSym).

init_pl(NumSym,(LVar,LVal)):-!,
	init_var_pl(NumSym,LVar,1),
	init_val_pl(NumSym,LVal,0).
init_pl(NumSym,LPerm):-
	init_point_pl(NumSym,LPerm,1).
	
init_point_pl(_,[],_).
init_point_pl(NumSym,[PointSym|LPointSym],Point):-
	init_point_c(NumSym,Point,PointSym),
	PointNxt is Point+1,
	init_point_pl(NumSym,LPointSym,PointNxt).

init_var_pl(_,[],_).
init_var_pl(NumSym,[IVarSym|LIVarSym],IVar):-
	init_var_c(NumSym,IVar,IVarSym),
	IVarNxt is IVar+1,
	init_var_pl(NumSym,LIVarSym,IVarNxt).

init_val_pl(_,[],_).
init_val_pl(NumSym,[ValSym|LValSym],Val):-
	init_val_c(NumSym,Val,ValSym),
	ValNxt is Val+1,
	init_val_pl(NumSym,LValSym,ValNxt).
	

% Initialisation des symtries de valeurs
init_sym_val(LPoint,Val1,Val2,Nb_var,Nb_val,1):-
	NbAff is Nb_var * Nb_val,
	length(LPoint,NbAff),
	enr_LVal(LPoint,1,Val1,Val2,Nb_val).
init_sym_val((LIVar,LVal),Val1,Val2,Nb_var,Nb_val,2):-
	length(LIVar,Nb_var),
	length(LVal,Nb_val),
	liste_num(LIVar,1,Nb_var),
	enr_LVal_val(LVal,1,Val1,Val2).


enr_LVal([],_,_,_,_):-!.
enr_LVal([XSym|LPoint],X,Val1,Val2,Nb_val):-
	I is (X-1)//Nb_val+1,
	J is (X-1) mod Nb_val,
	
	(J=Val1	->	JSym is Val2;
			(J=Val2	->	JSym is Val1;
					JSym=J)),
	XSym is (I-1)*Nb_val+JSym+1,
	Xp1 is X+1,
	enr_LVal(LPoint,Xp1,Val1,Val2,Nb_val).

enr_LVal_val([],_,_,_):-!.
enr_LVal_val([Val2|LVal],Val1,Val1,Val2):-!,
	Xp1 is Val1+1,
	enr_LVal_val(LVal,Xp1,Val1,Val2).
enr_LVal_val([Val1|LVal],Val2,Val1,Val2):-!,
	Xp1 is Val2+1,
	enr_LVal_val(LVal,Xp1,Val1,Val2).
enr_LVal_val([X|LVal],X,Val1,Val2):-!,
	Xp1 is X+1,
	enr_LVal_val(LVal,Xp1,Val1,Val2).
	
	
% Initialisation des symtries de valeurs sur un sous ensemble des variables
init_sym_val_local(LVar,Val1,Val2,Nb_var,Nb_val,LPoint):-
	NbAff is Nb_var * Nb_val,
	length(LPoint,NbAff),
	enr_LVal_permute(LVar,LPoint,Val1,Val2,Nb_val).

enr_LVal_permute([],LPoint,_,_,_):-combler(LPoint,1).
enr_LVal_permute([Var|LVar],LPoint,Val1,Val2,Nb_val):-
	X1 is (Var-1)*Nb_val+Val1+1,
	X2 is (Var-1)*Nb_val+Val2+1,
	nth(X1,LPoint,X2),
	nth(X2,LPoint,X1),
	enr_LVal_permute(LVar,LPoint,Val1,Val2,Nb_val).	

combler([],_):-!.
combler([Point|LPoint],I):-var(Point),!,
	Point=I,Ip1 is I+1,combler(LPoint,Ip1).
combler([_|LPoint],I):-
	Ip1 is I+1,combler(LPoint,Ip1).

% Initialisation des symtries de variables
init_sym_var(LPoint,Var1,Var2,Nb_var,Nb_val,1):-
	NbAff is Nb_var * Nb_val,
	length(LPoint,NbAff),
	enr_LVar(LPoint,1,Var1,Var2,Nb_val).
init_sym_var((LIVar,LVal),Var1,Var2,Nb_var,Nb_val,2):-
	length(LIVar,Nb_var),
	length(LVal,Nb_val),
	liste_num(LVal,1,Nb_val),
	enr_LIVar_var(LIVar,1,Var1,Var2),
	init_val_pl(NumSym,LIVar,1), init_var_pl(NumSym,LVal,1).


enr_LVar([],_,_,_,_).
enr_LVar([XSym|LPoint],X,Var1,Var2,Nb_val):-
	I is (X-1)//Nb_val+1,
	J is (X-1) mod Nb_val,
	
	(I=Var1	->	ISym is Var2;
			(I=Var2	->	ISym is Var1;
					ISym=I)),
	XSym is (ISym-1)*Nb_val+J+1,
	Xp1 is X+1,
	enr_LVar(LPoint,Xp1,Var1,Var2,Nb_val).

enr_LIVar_var([],_,_,_):-!.
enr_LIVar_var([IVar2|LIVar],IVar1,IVar1,IVar2):-!,
	Xp1 is IVar1+1,
	enr_LIVar_var(LIVar,Xp1,IVar1,IVar2).
enr_LIVar_var([IVar1|LIVar],IVar2,IVar1,IVar2):-!,
	Xp1 is IVar2+1,
	enr_LIVar_var(LIVar,Xp1,IVar1,IVar2).
enr_LIVar_var([X|LIVar],X,IVar1,IVar2):-!,
	Xp1 is X+1,
	enr_LIVar_var(LIVar,Xp1,IVar1,IVar2).

	
% Initialisation des symtries de variables sur un sous ensemble des valeurs
init_sym_var_local(LVar,Val1,Val2,Nb_var,Nb_val,LPoint):-
	NbAff is Nb_var * Nb_val,
	length(LPoint,NbAff),
	enr_LVar_permute(LVar,LPoint,Val1,Val2,Nb_val).

enr_LVar_permute([],LPoint,_,_,_):-combler(LPoint,1).
enr_LVar_permute([Val|LVal],LPoint,Var1,Var2,Nb_val):-
	X1 is (Var1-1)*Nb_val+Val+1,
	X2 is (Var2-1)*Nb_val+Val+1,
	nth(X1,LPoint,X2),
	nth(X2,LPoint,X1),
	enr_LVar_permute(LVal,LPoint,Var1,Var2,Nb_val).	

% Initialisation des symtries sous formes rduites
init_sym_red_form(LPoint,LVar,LVal,Nb_var,Nb_val,1):-
	NbAff is Nb_var * Nb_val,
	length(LPoint,NbAff),
	length(LVar,NbVarPerm),
	length(LVal,NbValPerm),
	enr_red_form(LPoint,1,LVar,LVal,Nb_val,NbVarPerm,NbValPerm).
init_sym_red_form((LVarSym,LValSym),LVar,LVal,Nb_var,Nb_val,2):-
	length(LVarSym,Nb_var),
	length(LValSym,Nb_val),
	enr_red_form2(LVarSym,1,LVar),
	enr_red_form2(LValSym,0,LVal).
	
enr_red_form([],_,_,_,_,_,_).
enr_red_form([XSym|LPoint],X,LVar,LVal,Nb_val,NbVarPerm,NbValPerm):-
	Var is (X-1)//Nb_val+1,
	Val is (X-1) mod Nb_val,

	belong_elt(Var,LVar,VarSym),
	belong_elt(Val,LVal,ValSym),

	XSym is (VarSym-1)*Nb_val+ValSym+1,
	Xp1 is X+1,
	enr_red_form(LPoint,Xp1,LVar,LVal,Nb_val,NbVarPerm,NbValPerm).
	
belong_elt(I,[],I).
belong_elt(I,[R|L],ISym):-
	(nth(Pos,R,I)	->
			(last(R,I)->NewPos=1;NewPos is Pos+1),
			nth(NewPos,R,ISym)
			;
			belong_elt(I,L,ISym)).
	
enr_red_form2([],_,_).
enr_red_form2([XSym|LPerm],X,LV):-

	belong_elt(X,LV,XSym),

	Xp1 is X+1,
	enr_red_form2(LPerm,Xp1,LV).

	
% init_sym_var_val

init_sym_var_val(_,_,_,_,_,1):-!.
init_sym_var_val((LVar,LVal),LVar,LVal,_,_,2):-!.
% max_dom(+LVar,0,-Max)

max_dom([],Max,Max).
max_dom([Var|LVar],I,Max):-
	fd_max(Var,MaxVar),
	(I>MaxVar->max_dom(LVar,I,Max);max_dom(LVar,MaxVar,Max)).

% convert_varval_to_point_list(-LPoint,+LVar,+LVal,+NbVar,+NbVal):
% convert a variable/value representation to a point representation
% output : list of permutation point LPoint
convert_varval_to_point_list(LPoint,LVar,LVal,NbVar,NbVal):-
	NbPoint is NbVar*NbVal,
	length(LPoint,NbPoint),
	convert_foreach_point(LPoint,1,LVar,LVal,NbVal).

convert_foreach_point([],_,_,_,_):-!.
convert_foreach_point([XSym|LPoint],X,LVar,LVal,NbVal):-
	Var is (X-1)//NbVal+1,
	Val is (X-1) mod NbVal,
	Valp1 is Val+1,
	
	nth(Var,LVar,VarSym),
	nth(Valp1,LVal,ValSym),

	XSym is (VarSym-1)*NbVal+ValSym+1,
	Xp1 is X+1,
	convert_foreach_point(LPoint,Xp1,LVar,LVal,NbVal).

% manipulation des listes de permutations
paste_var_val(LVar,[],_,LVar).
paste_var_val(LVar,[Val|LVal],N,LPoint2):-
	paste_var_val(LVar,LVal,N,LPoint1),
	Point is Val+N+1,
	append(LPoint1,[Point],LPoint2).

add_to_list([],_,[]).
add_to_list([P1|LIn],N,[P2|LOut]):-
	P2 is P1+N,
	add_to_list(LIn,N,LOut).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Fonctions de recouvrement des symtries		%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% recover_solutions(+LSol,-LSolTotal,+LSym,+Param):

% retrouve l'ensemble total LSolTotal des solutions 
% partir de l'ensemble LSol des solutions trouves compos
% avec les symtries de LSym.


% sol_sym(+Sol,+Sym,-SolSym,+Param,+LastIndex,+SolSize):
% calcule la symtrie SolSym de Sol selon la fonction Sym.
	
sol_sym([Val],Sym,Sol_Sym,IVar):-!,
	transposition(IVar,Val,IVarSym,ValSym,Sym),
	nth(IVarSym,Sol_Sym,ValSym).	
sol_sym([Val|Sol],Sym,Sol_Sym,IVar):-
	IVar_nxt is IVar-1,
	sol_sym(Sol,Sym,Sol_Sym,IVar_nxt),
	transposition(IVar,Val,IVarSym,ValSym,Sym),
	nth(IVarSym,Sol_Sym,ValSym).


% transposition(+IVar,+Val,-IVarSym,-ValSym,+Sym,+Param,+N):
% calcule la symtrie (IVarSym,ValSym) du couple (IVar,Val)
% selon la symtrie Sym.
transposition(IVar,Val,IVarSym,ValSym,Sym):-
	(image_sym(Sym,(IVar,Val),(IVarSym_tmp,ValSym_tmp))->
		IVarSym=IVarSym_tmp,ValSym=ValSym_tmp
		;
		IVarSym=IVar,ValSym=Val
	).

check_sym_br(Sym) :-
	findall(Sol,queen_sbds(Sol,4),LSol),!,
	length(LSol,NbSol),
	write('nombre de sol: '),write(NbSol),nl,
	member(Sol1,LSol),reverse(Sol1,Sol1r),
	sol_sym(Sol1r,Sym,Sol2,4),
	(member(Sol2,LSol)->
		write(Sol1),nl,
		write(Sol2),nl;
		fail).
	
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Fonctions de symtrie			%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Fonctions de construction de symtrie

	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Fonction test


incr :-
	g_read(cont,Cont),Cont2 is Cont+1,g_assign(cont,Cont2).

compt(Pred) :- 
	g_assign(cont,0),call(Pred),incr,g_read(cont,Cont),write(Cont),nl,fail.


